"""Library to load new image on a FTD 5500x. The applies to the Kenton
platforms (5506, 5508, 5516), Saleen platforms (5512, 5525, 5545, 5555).

We require a file server to be co-located with the device. This file
server has both the rommon_image (lfbff for Kenton, and cdisk for Saleen)
and the pkg_image mounted. FTD will first break into rommon, tftp the
rommon_image and boot from it into FTD boot CLI. Then it will do an HTTP
copy of pkg_image to boot up into a fully functional device. This second
step takes a long time (~30 min), as it will run a lot of scripts.

The file server is a PXE server in each site, defined in gcp_constants.py.

"""
import logging
import os.path
import re
import time

from kick.miscellaneous.credentials import *

KICK_EXTERNAL = False

try:
    from kick.file_servers.file_servers import *
except ImportError:
    KICK_EXTERNAL = True

logger = logging.getLogger(__name__)

from unicon.eal.expect import TimeoutError
from unicon.eal.dialogs import Dialog
from unicon.eal.utils import ExpectMatch
from pathlib import PurePosixPath
from ...general.actions.basic import BasicDevice, BasicLine
from kick.device2.general.actions.power_bar import power_cycle_all_ports

try:
    import kick.graphite.graphite as graphite
except ImportError:
    import kick.metrics.metrics as graphite

MAX_RETRY_COUNT = 3
DEFAULT = 'DEFAULT'
EN_PASSWORD = 'en_password'
DEFAULT_ENPASSWORD = 'myenpassword'


class Ftd5500x(BasicDevice):
    def __init__(self, hostname='firepower', login_password='Admin123',
                 sudo_password='Admin123', enable_password='', *args, **kwargs):
        """Constructor for FTD.

        Add *args and **kwargs so that the initiator can be invoked with
        additional (unused) arguments.

        :param hostname: host name as appears in prompt
                e.g. 'BATIT-5508-3-FUL'
        :param login_password: password for login
        :param sudo_password: root password
        :param enable_password: password for privileged exec mode
        :return: None

        """

        super().__init__()

        graphite.publish_kick_metric('device.ftd5500x.init', 1)
        from .patterns import Ftd5500xPatterns
        self.patterns = Ftd5500xPatterns(hostname, login_password, sudo_password, enable_password)

        # create the state machine that contains the proper attributes.
        from .statemachine import Ftd5500xStatemachine
        self.sm = Ftd5500xStatemachine(self.patterns)

        # important: set self.line_class so that a proper line can be created
        # by ssh_console(), etc.
        self.line_class = Ftd5500xLine

    def clear_line(self, line, port, en_password=DEFAULT_ENPASSWORD):
        """Function to clear line.

        :param line: spawn ID for the line of terminal server connection
        :param port: device line port in terminal server to be cleared
        :param en_password: enable password to switch to line configuration mode
        :return: None

        """
        if en_password == DEFAULT_ENPASSWORD:
            en_password = get_password(en_password)

        try:
            line.spawn_id.expect('#')
        except:
            # expect >
            line.spawn_id.sendline('en')
            try:
                line.spawn_id.expect('Password:')
                line.spawn_id.sendline(en_password)
            except:
                pass
        #port = port[-2:]
        #if port[-2] is '0':
        #    port = port[-1]
        line.spawn_id.sendline('clear line {}'.format(port))
        line.spawn_id.expect('[confirm]')
        line.spawn_id.sendline('')
        line.spawn_id.expect('[OK]')
        logger.info('line: {} cleared'.format(port))

    def check_version(self, ftd_line, version, build):
        """Checks if the installed version matches a certain version.

        :param ftd_line: Instance of FTD line used to connect to FTD
        :param version: version to be checked
        :param build: build to be checked
        :return: None

        """
        # ftd_line = self.telnet_console_with_credential(ip, port)
        ftd_line.go_to('any')
        ftd_line.go_to('fireos_state')

        # check version
        response = ftd_line.execute('show version', 30)
        current_build = re.findall('Build\s(\d+)', response)[0]
        current_version = re.findall(r'(Version\s){1}([0-9.]+\d)', str(response))[0][1]
        if current_build == build and current_version == version:
            logger.info('the same version and build')
        else:
            logger.error('not the same version and build')

    def log_checks(self, ftd_line, list_files=['/var/log/messages'],
                   search_strings=['fatal', 'error'], exclude_strings=[], timeout=300):
        """Wrapper function to Get logs for FTD.

        :param ftd_line: Instance of FTD line used to connect to FTD
        :param list_files: List of file paths for files to search in
        :param search_strings: List of keywords to loook for in the logs
        :param exclude_strings: List of keywords to be excluded in the logs
        :param timeout: time to wait for log retrieval in seconds

        e.g.:
            list_files = ['/var/log/boot_1341232','/var/log/boot_*'] \n
            search_strings = ['fatal','error', 'crash'] \n
            exclude_strings = ['ssl_flow_errors', 'firstboot.S09']

        """

        self.sm.go_to('sudo_state', ftd_line.spawn_id)

        grep_command_list = []
        exclude_line = ''

        if exclude_strings:
            exclude_cmd = ['| grep -v {}'.format(string) for string in exclude_strings]
            exclude_line = ''.join(exclude_cmd)

        if list_files and search_strings:
            for file in list_files:
                for string in search_strings:
                    grep_command = "grep -Ii {} {} | " \
                                   "sort -u {}".format(string, file, exclude_line)
                    grep_command_list.append(grep_command)
        try:
            output_log = ftd_line.execute_lines(("\n".join(grep_command_list)),
                                                timeout=timeout)
        except:
            output_log = "Log retrieval command timed out"

        logger.info("""
            ***********************************************************

            Logs for the requested files in FTD are : -

            {}

            ***********************************************************
            """.format(output_log))
  
    def ssh_vty(self, ip, port, username='admin', password='Admin123',
                timeout=None, line_type='ssh_vty', rsa_key=None):
        return super().ssh_vty(ip, port, username=username, password=password,
                               timeout=timeout, line_type=line_type, rsa_key=rsa_key)


class Ftd5500xLine(BasicLine):
    def __init__(self, spawn, sm, type, timeout=None):
        """ Constructor of Ftd5500x

        :param spawn: spawn_id of the line
        :param sm: object of statemachine
        :param type: line type, e.g. 'ssh', 'telnet'
        :return: object of the line

        """

        try:
            super().__init__(spawn, sm, type, timeout)
        except:
            if self.type is 'ssh_vty':
                raise
            # Handle boot loop
            logger.info("Try to drop device to rommon state.")
            logger.info('=== Failed to go_to("any") '
                        'try to drop device to rommon.')
            try:
                self.rommon_go_to(timeout=300)
                self.sm.update_cur_state('rommon_state')
                logger.info('=== Dropped device to rommon.')
            except:
                logger.info("Failed to go to rommon. Unknown device state.")

    def expert_execute(self, cmd, timeout=None, exception_on_bad_command=False):
        """Execute a command in expert mode

        :param cmd: command to be executed in expert state given as a string
        :param timeout: in seconds
        :param exception_on_bad_command: True/False - whether to raise an exception
                on a bad command
        :return: output as string

        """

        self.go_to('expert_state')
        return self.execute(cmd, timeout, exception_on_bad_command)

    def enable_execute(self, cmd, timeout=None, exception_on_bad_command=False):
        """Go to enable mode, run the command and return the output.

        :param cmd: command to be executed given as a string
        :param timeout: in seconds
        :param exception_on_bad_command: True/False - whether to raise an exception
            on a bad command
        :return: output as string

        """

        # first, issue "no terminal pager"
        # then run command
        super().enable_execute('no terminal pager')
        return super().enable_execute(cmd, timeout, exception_on_bad_command)

    def wait_until_device_on(self, timeout=600):
        """ Wait a given period of time for device to be on

        :param timeout: wait time for the device to boot up
        :return: None

        """

        # The system will reboot, wait for the following prompts
        d1 = Dialog([
            ['SW-DRBG health test passed', None, None, False, False],
        ])
        d1.process(self.spawn_id, timeout=timeout)

        # sleep 60 seconds to avoid errors
        time.sleep(60)
        self.sm.go_to('any', self.spawn_id)

    def power_cycle(self, power_bar_server, power_bar_port,
                    wait_until_device_is_on=True, timeout=600,
                    power_bar_user='admn', power_bar_pwd='admn'):
        """reboots a device from a Power Data Unit equipment.

        :param power_bar_server: comma-separated string string of IP addresses of the PDU's
        :param power_bar_port: comma-separated string string of power ports on the PDU's
        :param wait_until_device_is_on: True if waits for the device
                                to boot up, False otherwise
        :param timeout: wait time for the device to boot up
        :param power_bar_user: comma-separated string of usernames for power bar servers
        :param power_bar_pwd: comma-separated string of passwords for power bar servers
        :return: status of power cycle result, True or False

        """

        if power_bar_server == "" or power_bar_port == "":
            return None

        result = power_cycle_all_ports(power_bar_server, power_bar_port, power_bar_user, power_bar_pwd)

        if wait_until_device_is_on:
            logger.info('Wait for device to be up and running ...')
            self.wait_until_device_on(timeout=timeout)

        return result

    # the following functions are for device baselining. This only works on
    # a console line (ssh or telnet).
    def expect_and_sendline(self, this_spawn, es_list, timeout=10):
        """takes a list of expect/send actions and perform them one by one.

        es_list looks like:
        [['exp_pattern1', 'send_string1', 30],
         ['exp_pattern2', 'send_string2'],
         ...
        ]

        The third element is for timeout value, and can be ommitted when the
        overall timeout value applies.

        :param this_spawn: the spawn of the line
        :param es_list: pairs of expected output and command to be sent
        :param timeout: default timeout for the command if timeout is not set
        :return: None

        """

        for es in es_list:
            if len(es) == 2:
                exp_pattern = es[0]
                send_string = es[1]
                to = timeout
            elif len(es) == 3:
                exp_pattern = es[0]
                send_string = es[1]
                to = int(es[2])
            else:
                raise RuntimeError("Unknown expect_and sendline input")

            this_spawn.sendline(send_string)
            this_spawn.expect(exp_pattern, timeout=to)
            time.sleep(3)

    def rommon_go_to(self, timeout=300):
        """Go to rommon mode.

        :return: None

        """

        # break into rommon
        d1 = Dialog([
            ['Use SPACE to begin boot immediately\.',
             'sendline({})'.format(chr(27)), None, True, False],
            ['Boot in \d+ seconds', 'sendline({})'.format(chr(27)), None,
             True, False],
            ['Boot interrupted\.', None, None, False, False]
        ])
        d1.process(self.spawn_id, timeout=timeout)
        time.sleep(10)
        logger.info('In rommon mode')

    def rommon_configure(self):
        """In ROMMON mode, set network, tftp info and rommon image For Kenton.

        :return: None

        """

        logger.info('add rommon config')
        es_list = [
            ['rommon', 'ADDRESS={}'.format(self.uut_ip)],
            ['rommon', 'NETMASK={}'.format(self.uut_netmask)],
            ['rommon', 'SERVER={}'.format(self.rommon_tftp_server)],
            ['rommon', 'IMAGE={}'.format(self.rommon_image)],
            ['rommon', 'GATEWAY={}'.format(self.uut_gateway)],
            ['rommon', 'PORT=Management0/0'],
        ]

        self.expect_and_sendline(self.spawn_id, es_list)

    def rommon_configure_saleen(self):
        """In ROMMON mode, set network, tftp info and rommon image For Saleen.

        :return: None

        """

        logger.info('add rommon config')
        es_list = [
            ['rommon', 'address {}'.format(self.uut_ip)],
            ['rommon', 'gateway {}'.format(self.uut_gateway)],
            ['rommon', 'server {}'.format(self.rommon_tftp_server)],
            ['rommon', 'file {}'.format(self.rommon_image)],
        ]
        self.expect_and_sendline(self.spawn_id, es_list)

    def rommon_boot(self, timeout=600):
        """Perform TFTP Boot from ROMMON.

        :return: None

        """

        time.sleep(10)
        logger.info('=== Performing TFTP Boot from ROMMON')
        logger.info('=== Wait for installation to complete, timeout = {}'
                    ' seconds ...'.format(str(timeout)))
        self.spawn_id.sendline('tftpdnld')
        d1 = Dialog([
            ['rommon.*> ', 'sendline(tftpdnld)', None, True, False],
            ['File not found', None, None, False, False],
            ['Use SPACE to launch Cisco FTD immediately.',
             'sendline({})'.format(chr(27)), None, True, False],
            ['-boot>', 'sendline()', None, False, False],
            [self.sm.patterns.prompt.enable_prompt, 'sendline()', None, False, False],
            [self.sm.patterns.prompt.disable_prompt, 'sendline()', None, False, False],
        ])
        try:
            response = d1.process(self.spawn_id, timeout=timeout)
            self.spawn_id.sendline()
            logger.info("=== Rommon file was installed successfully.")
        except:
            logger.info("=== Rommon file download failed, raise runtime error. ")
            raise RuntimeError(
                ">>>>>> Download failed. Please check details - "
                "tftp_server: {}, image file: {}".format(self.rommon_tftp_server,
                                                         self.rommon_image))
        if 'File not found' in response.match_output:
            raise FileNotFoundError("TFTP error: File not found")

        logger.info('In boot CLI mode')
        logger.info("=== Rommon file was installed successfully.")

    def firepower_boot_configure(self):
        """In boot CLI mode, set device network info, DNS server and domain.

        :return: None

        """

        d1 = Dialog([
            ['-boot>', 'sendline(setup)', None, False, False],
        ])
        d1.process(self.spawn_id)
        set_ntp_server = 'n' if self.ntp_server == None else 'y'
        d2 = Dialog([
            ['Enter a hostname ', 'sendline(firepower)', None, True, False],
            ['Do you want to configure IPv4 address on management interface',
             'sendline(y)', None, True, False],
            ['Do you want to enable DHCP for IPv4 address assignment on '
             'management interface', 'sendline(n)', None, True, False],
            ['Enter an IPv4 address', 'sendline({})'.format(self.uut_ip),
             None, True, False],
            ['Enter the netmask', 'sendline({})'.format(self.uut_netmask),
             None, True, False],
            ['Enter the gateway', 'sendline({})'.format(self.uut_gateway),
             None, True, False],
            ['Do you want to configure static IPv6 address on management '
             'interface', 'sendline(n)', None, True, False],
            ['Enter the primary DNS server IP address',
             'sendline({})'.format(self.dns_server.split(',')[0]), None, True, False],
            ['Do you want to configure Secondary DNS Server', 'sendline(n)',
             None, True, False],
            ['Do you want to configure Local Domain Name', 'sendline(n)',
             None, True, False],
            ['Do you want to configure Search domains', 'sendline(y)', None,
             True, False],
            ['Enter the comma separated list for search domains',
             'sendline({})'.format(self.search_domains), None, True, False],
            ['Do you want to enable the NTP service', 'sendline({})'.format(set_ntp_server), None,
             True, False],
            ['Enter the NTP servers separated by commas', 'sendline({})'.format(self.ntp_server), None,
             True, False],
            ['Apply the changes', 'sendline(y)', None, True, False],
            ['Press ENTER to continue...', 'sendline()', None, True, False],
            ['-boot>', 'sendline()', None, False, False],
        ])
        d2.process(self.spawn_id, timeout=120)

    def _last_match_index(self, match):
        """
        Makes the next, firepower_install easier to run unittest.
        :param match:
        :return:
        """
        return match.last_match_index

    def firepower_install(self):
        """Perform ping test and verify the network connectivity to TFTP server.
        Install FTD pkg image Enter device network info, hostname, and firewall
        mode.

        :return: None

        """

        for i in range(40):
            self.spawn_id.sendline('ping -c 1 {}'.format(self.rommon_tftp_server))
            try:
                self.spawn_id.expect('64 bytes from', timeout=5)
            except TimeoutError:
                time.sleep(60)
                continue
            else:
                break
        else:
            raise RuntimeError(">>>>>> Ping not working")

        d0 = Dialog([
            ['-boot>', 'sendline(system install {})'.format(self.pkg_image),
                None, False, False]
        ])

        d1 = Dialog([
            ['Do you want to continue\?', 'sendline(y)', None, True, False],
            ['Upgrade aborted', 'sendline()', None, False, False],
            ['Installation aborted', 'sendline()', None, False, False],
            ['Package Detail', None, None, False, False]
        ])
        count = 0
        while count < self.retry_count:
            d0.process(self.spawn_id, timeout=20)
            match = d1.process(self.spawn_id, timeout=600)
            if self._last_match_index(match) == 3:
                break
            count += 1
            time.sleep(5)
        else:  # didn't break out
            raise RuntimeError('ftd installation failed, please check '
                'ftd package url: "{}"'.format(self.pkg_image))

        d2 = Dialog([
            ['Do you want to continue with upgrade\?', 'sendline(y)', None,
             False, False],
        ])
        # used to take 2 minutes
        d2.process(self.spawn_id, timeout=1200)

        # Handle the case "Press 'Enter' to reboot the system." was never displayed
        # Script stuck if message above is not displayed but the device is waiting for 'enter'
        count = 0
        waittime = 60
        total = 60 
        d2_2 = Dialog([
            ["Press 'Enter' to reboot the system.", 'sendline()', None, True,
             False],
            ['Use SPACE to begin boot immediately.', 'send(" ")', None, True,
             False],
            ['Use SPACE to launch Cisco FTD immediately.', 'send(" ")', None,
             True, False],
            ['firepower login: ', 'sendline()', None, False, False],
        ])

        while count < total:
            self.spawn_id.sendline()
            try: 
                d2_2.process(self.spawn_id, timeout=waittime)
                break
            except:
                logger.info('=== Wait for firepower login: ({})'.format(count))
                count += 1
                time.sleep(5)
                continue

        # Allow install processes to finish
        # Sleep was extended to 5 minutes as part of CSCvi89671
        # TODO: This sleep should be removed once CSCvi89616 is resolved
        time.sleep(300)

        d3 = Dialog([
            ['firepower login: ', 'sendline(admin)', None, True, False],
            ['Password: ', 'sendline({})'.format(self.sm.patterns.default_password),
             None, True, False],
            ['Press <ENTER> to display the EULA: ', 'sendline()', None, True, False],
            ['--More--', 'send(q)', None, True, False],
            ["Please enter 'YES' or press <ENTER> to AGREE to the EULA: ",
             'sendline(YES)', None, False, False],
        ])
        d3.process(self.spawn_id, timeout=600)

        d4 = Dialog([
            ['firepower login: ', 'sendline(admin)', None, True, False],
            ['Password: ', 'sendline({})'.format(self.sm.patterns.login_password),
             None, True, False],
            ['Enter new password:', 'sendline({})'.format(self.sm.patterns.login_password),
             None, True, True],
            ['Confirm new password:', 'sendline({})'.format(self.sm.patterns.login_password),
             None, True, False],
            ['Do you want to configure IPv4', 'sendline(y)', None, True, False],
        ])

        if self.uut_ip6 is None:
            d4.append(['Do you want to configure IPv6', 'sendline(n)', None, True, False])
        else:
            d4.append(['Do you want to configure IPv6', 'sendline(y)', None, True, False])
        d4.append(['Configure IPv4 via DHCP or manually', 'sendline(manual)', None,
                   True, False])
        d4.append(['Enter an IPv4 address for the management interface',
                   'sendline({})'.format(self.uut_ip), None, True, False])
        d4.append(['Enter an IPv4 netmask for the management interface',
                   'sendline({})'.format(self.uut_netmask), None, True, False])
        d4.append(['Enter the IPv4 default gateway for the management interface',
                   'sendline({})'.format(self.uut_gateway), None, True, False])
        if self.uut_ip6 is not None:
            d4.append(['Configure IPv6 via DHCP, router, or manually',
                       'sendline(manual)', None, True, False])
            d4.append(['Enter the IPv6 address for the management interface',
                       'sendline({})'.format(self.uut_ip6), None, True, False])
            d4.append(['Enter the IPv6 address prefix for the management interface',
                       'sendline({})'.format(self.uut_prefix), None, True, False])
            d4.append(['Enter the IPv6 gateway for the management interface',
                       'sendline({})'.format(self.uut_gateway6), None, True, False])
        d4.append(['Enter a fully qualified hostname for this system',
                   'sendline({})'.format(self.hostname), None, True, False])
        d4.append(['Enter a comma-separated list of DNS servers or',
                   'sendline({})'.format(self.dns_server), None, True, False])
        d4.append(['Enter a comma-separated list of search domains or',
                   'sendline({})'.format(self.search_domains), None, False, False])
        d4.process(self.spawn_id, timeout=900)

        d5 = Dialog([
            ['Configure (firewall|deployment) mode', 'sendline({})'.format(self.firewall_mode),
             None, True, False]
        ])

        if self.mode == 'local':
            d5.append(['Manage the device locally?', 'sendline(yes)', None, True, False])
        else:
            d5.append(['Manage the device locally?', 'sendline(no)', None, True, False])
        d5.append(['Successfully performed firstboot initial configuration steps',
                   'sendline()', None, True, False])
        d5.append(['> ', 'sendline()', None, False, False])
        d5.process(self.spawn_id, timeout=600)

        logger.info('fully installed.')

    def configure_manager(self):
        """Configure manager to be used for registration

        :return: None

        """

        self.go_to('fireos_state')
        if self.manager_nat_id is None:
            response = self.execute('configure manager add {} {}'.format(self.manager,
                                                                         self.manager_key),
                                    120)
        else:
            response = self.execute('configure manager add {} {} {}'.format(self.manager,
                                                                            self.manager_key,
                                                                            self.manager_nat_id),
                                    120)
        success = re.search('Manager successfully configured', response)
        if success is None:
            logger.error('Exception: failed to configure the manager')
            raise RuntimeError('>>>>>> configure manager failed:\n{}\n'.format(response))

    def validate_version(self):
        """Checks if the installed version matches the version in package url

        :return: None

        """

        self.go_to('fireos_state')

        # check version
        response = self.execute('show version', 30)
        if not response:
            response = self.execute('show version', 30)
        build = re.findall('Build\s(\d+)', response)[0]
        version = re.findall(r'(Version\s){1}([0-9.]+\d)', str(response))[0][1]
        ftd_package_file = self.pkg_image.split('/')[-1]
        if (build in ftd_package_file) and (version in ftd_package_file):
            logger.info('>>>>>> show version result:\n{}\nmatches '
                        'ftd package image: {}'.format(response,
                                                       ftd_package_file))
            logger.info('Installed ftd version validated')
        else:
            logger.error('Exception: not the same version and build')
            raise RuntimeError('>>>>>> show version result:\n{}\ndoes not match '
                               'ftd package image: {}'.format(response,
                                                              ftd_package_file))

    def rommon_to_new_image(self, rommon_tftp_server, pkg_image,
                            uut_ip, uut_netmask, uut_gateway, rommon_image, dns_server,
                            hostname='firepower', search_domains='cisco.com',
                            is_device_kenton=True, retry_count=MAX_RETRY_COUNT,
                            power_cycle_flag=False, pdu_ip='', pdu_port='',
                            pdu_user='admn', pdu_pwd='admn', ntp_server=None,
                            mode='local',
                            uut_ip6=None, uut_prefix=None, uut_gateway6=None,
                            manager=None, manager_key=None, manager_nat_id=None,
                            firewall_mode='routed', timeout=600, **kwargs):
        """Install rommon image and FTD pkg image on ASA.

        :param rommon_tftp_server: TFTP Server IP Address
        :param pkg_image: FTD image to be transferred via HTTP,
            e.g. 'http://192.168.0.50/Release/6.0.0-1005/installers/ftd-6.0.0-1005.pkg'
        :param uut_ip: Device IP Address to access TFTP Server
        :param uut_netmask: Device Netmask
        :param uut_gateway: Device Gateway
        :param rommon_image: boot image under /tftpboot to be transferred via TFTP,
            e.g. 'asa/Release/6.0.0-1005/installers/ftd-boot-99.1.3.194.lfbff'
        :param dns_server: DNS server
        :param hostname: hostname to be set
        :param search_domains: search domains delimited by comma,
            defaulted to 'cisco.com'
        :param is_device_kenton: True if device is Kenton, False if device is Saleen
        :param retry_count: download retry count, defaulted to MAX_RETRY_COUNT
        :param power_cycle_flag: True power cylce before baseline, False otherwise
        :param pdu_ip: PDU IP
        :param pdu_port: PDU Port
        :param pdu_user: PDU admn
        :param pdu_pwd: PDU pwd
        :param ntp_server: NTP server delimited by comma, defaulted to None,
            otherwise the value of ntp_server is - e.g. "ntp.esl.cisco.com"
        :param mode: the manager mode (local, remote)
        :param uut_ip6: Device IPv6 Address
        :param uut_prefix: Device IPv6 Prefix
        :param uut_gateway6: Device IPv6 Gateway
        :param manager: FMC to be configured for registration
        :param manager_key: Registration key
        :param manager_nat_id: Registration NAT Id
        :param firewall_mode: the firewall mode (routed, transparent, ngips)
        :param timeout: in seconds; time to wait for fetching the boot image from TFTP server;
                        defaulted to 600s
        :return: None

        """

        logger.info('Starting baseline')
        graphite.publish_kick_metric('device.ftd5500x.baseline', 1)
        self.rommon_tftp_server = rommon_tftp_server
        self.pkg_image = pkg_image
        self.uut_ip = uut_ip
        self.uut_netmask = uut_netmask
        self.uut_gateway = uut_gateway
        self.uut_ip6 = uut_ip6
        self.uut_prefix = uut_prefix
        self.uut_gateway6 = uut_gateway6
        self.rommon_image = rommon_image
        self.dns_server = dns_server
        self.hostname = hostname
        self.search_domains = search_domains
        self.retry_count = retry_count
        self.ntp_server = ntp_server
        self.mode = mode
        self.firewall_mode=firewall_mode
        self.manager = manager
        self.manager_key = manager_key
        self.manager_nat_id = manager_nat_id
        if not (self.sm.current_state in ['rommon_state', 'boot_state']):
            if not power_cycle_flag:
                if self.sm.current_state is 'disable_state':
                    logger.info('Device is in disable state. Go to enable and reload ...')
                    self.go_to('enable_state')
                    self.spawn_id.sendline('reload noconfirm')
                elif self.sm.current_state in ['enable_state', 'config_state']:
                    logger.info('Device is in enable or config state. Reloading ...')
                    self.spawn_id.sendline('reload noconfirm')
                else:
                    logger.info('Reboot the device ...')
                    self.go_to('sudo_state')
                    self.spawn_id.sendline('reboot')
                    try:
                        self.spawn_id.expect('Rebooting...', timeout=300)
                    except TimeoutError:
                        raise RuntimeError(">>>>>> Failed to reboot the device. Probably hanged during reboot?")
            else:
                logger.info('Power cycle the device ...')
                self.power_cycle(pdu_ip, pdu_port, wait_until_device_is_on=False, power_bar_user=pdu_user,
                                 power_bar_pwd=pdu_pwd)
            logger.info('Drop the device to rommon.')
            self.rommon_go_to()

        if self.sm.current_state is 'boot_state':
            logger.info('Device is in boot_state, drop the device to rommon.')
            self.spawn_id.sendline('system reload')
            d1 = Dialog([
                ['Are you sure you want to reload the system',
                 'sendline(y)', None, False, False],
            ])
            d1.process(self.spawn_id, timeout=30)
            self.rommon_go_to()

        self.rommon_config(is_device_kenton)

        logger.info('tftpdnld - tftp server: {}, '
                    'rommon image: {} ...'.format(rommon_tftp_server, rommon_image))
        try:
            self.rommon_boot(timeout=timeout)
        except FileNotFoundError:
            logger.info('TFTP Error: File not found! Run again the network settings')
            self.rommon_config(is_device_kenton)
            self.rommon_boot(timeout=timeout)

        self.go_to('any')
        logger.info('firepower boot configure ...')
        self.firepower_boot_configure()
        logger.info('FTD image install - image: {} ...'.format(pkg_image))
        self.firepower_install()
        self.go_to('any')
        if self.manager is not None and self.mode != 'local':
            logger.info('Configure manager')
            self.configure_manager()
        logger.info('Validate version installed')
        self.validate_version()
        logger.info('Installation completed successfully.')

    def rommon_config(self, is_device_kenton):
        if is_device_kenton:
            logger.info('Device is Kenton. Rommon configure.')
            self.rommon_configure()
        else:
            logger.info('Device is Saleen. Rommon configure.')
            self.rommon_configure_saleen()

    def enable_configure(self, timeout=20):
        """Used to enable configure terminal

        :param timeout: time to wait for the output of a command;
                        default is 30s
        :return: None

        """
        d = Dialog([
            ['-boot>', 'sendline()', None, False, False],
            [self.sm.patterns.prompt.enable_prompt, 'sendline()', None, False, False],
            [self.sm.patterns.prompt.disable_prompt, 'sendline()', None, False, False],
            ['login', 'sendline({})'.format(self.sm.patterns.username), None, True, False],
            ['(p|P)assword:', 'sendline({})'.format(self.sm.patterns.login_password), None, False, False],
        ])

        d1 = Dialog([
            ['Would you like to enable anonymous error reporting to help improve', 'sendline(N)',
             None, True, False],
            [self.sm.patterns.prompt.config_prompt, 'sendline()', None, False, False],
        ])

        try:
            self.go_to('any')
        except:
            logger.info("Device is rebooting")
            d.process(self.spawn_id, timeout=600)
            self.go_to('any')

        self.go_to('enable_state')
        time.sleep(0.5)
        logger.info("=== Configure terminal ")
        self.spawn_id.sendline('conf t')
        time.sleep(0.5)
        if 'Invalid input detected' in self.spawn_id.read():
            self.spawn_id.sendline()
            time.sleep(0.5)
            self.spawn_id.expect(self.sm.patterns.prompt.enable_prompt)
            output = self.execute('show running-config | grep Serial Number', 15)
            sn = re.findall(r'Serial Number:\s(\w+)', output)[0]
            new_sn = '1111222233334444' + sn
            self.sudo_execute('echo -n {}| md5sum > /mnt/disk0/enable_configure'.format(new_sn), timeout)
            self.go_to('config_state')
            d1.process(self.spawn_id, timeout=30)

    def set_boot_image(self, tftp_server, boot_image, timeout=20):
        """Function for setting the new boot system image

        :param tftp_server: tftp server ip
                e.g: '192.168.0.50'
        :param boot_image: the name of the boot image given with its absolute path
               e.g. asa/asa952-2-smp-k8.bin for saleen or asa/asa962-21-lfbff-k8.SPA for Kenton
               asa962-21-lfbff-k8.SPA
        :param timeout: defaulted to 20s
        :return: None

        """

        self.go_to('config_state')
        # remove previous images
        output = self.execute('show running-config boot system')
        if output:
            images = re.findall(r'disk0:/.*', output)
            logger.info('Removing previous boot system images')
            if images:
                for i in images:
                    self.execute('no boot system {}'.format(i), timeout)

        logger.info('Copying the new ASA image to the ASA, placing the image in disk0')
        asa_boot_image = PurePosixPath(boot_image).name or boot_image
        command = 'copy /noconfirm tftp://{}/{} disk0:{}'.format(tftp_server, boot_image, asa_boot_image)
        self.spawn_id.sendline(command)
        d1 = Dialog([
            ['Do you want to over write?', 'sendline()', None, True, False],
            ['Accessing tftp', None, None, True, True],
            ['Digital signature successfully validated', None, None, True, False],
            ['bytes copied in .* secs', 'sendline()', None, False, False],
            ['Error', None, None, False, False],
            [self.sm.patterns.prompt.config_prompt, 'sendline()', None, False, False],
        ])
        resp = d1.process(self.spawn_id, timeout=1200)
        if isinstance(resp, ExpectMatch):
            if 'Error' in resp.match_output:
                logger.error('An error appeared while downloading the image, '
                             'please check the name of the boot image and its path')
                raise RuntimeError('An error appeared while downloading the boot image, '
                                   'please check your file path')

        logger.info('Setting new boot system image')
        self.execute('boot system disk0:/{}'.format(asa_boot_image, timeout))
        self.spawn_id.sendline('write memory')
        d2 = Dialog([
            #[self.sm.patterns.prompt.config_prompt, 'sendline(write memory)', None, True, False],
            ['Building configuration', None, None, True, False],
            ['[OK]', 'sendline()', None, True, False],
            [self.sm.patterns.prompt.config_prompt, None, None, False, False],
        ])
        d2.process(self.spawn_id, timeout=60)

    def configure_mgmt_interface(self, ip, netmask, gateway, tftp_server, mgmt_port='Management1/1'):
        """

        :param ip: ip address of the management interface
        :param netmask: netmask of the network
        :param gateway: gateway ip address
        :param tftp_server: ip address of tftp server
        :param mgmt_port: name of the management port; if not provided, defaulted to 'Management1/1'
        :return: None
        """

        self.go_to('config_state')
        es_list = [
            [self.sm.patterns.prompt.config_prompt, 'conf t'],
            [self.sm.patterns.prompt.config_prompt, 'int {}'.format(mgmt_port)],
            [self.sm.patterns.prompt.config_prompt, 'ip address {} {}'.format(ip, netmask)],
            [self.sm.patterns.prompt.config_prompt, 'nameif diagnostic'],
            [self.sm.patterns.prompt.config_prompt, 'no shutdown'],
            [self.sm.patterns.prompt.config_prompt, 'no cts manual'],
            [self.sm.patterns.prompt.config_prompt, 'security-level 100'],
            [self.sm.patterns.prompt.config_prompt, 'exit'],
            [self.sm.patterns.prompt.config_prompt, 'route diagnostic 0 0 {} 1'.format(gateway)],
            [self.sm.patterns.prompt.config_prompt, 'ping {}'.format(tftp_server)],
        ]
        self.expect_and_sendline(self.spawn_id, es_list, 20)

    def convert_to_elektra(self, rommon_tftp_server,
        uut_ip, uut_netmask, uut_gateway, asa_image, dns_server,
        hostname='firepower', search_domains='cisco.com',
        is_device_kenton=True, retry_count=MAX_RETRY_COUNT,
        power_cycle_flag=False, pdu_ip='', pdu_port='', mgmt_port='Management1/1', timeout=600):
        """Install rommon image and FTD pkg image on ASA.

        :param rommon_tftp_server: TFTP Server IP Address
        :param uut_ip: Device IP Address to access TFTP Server
        :param uut_netmask: Device Netmask
        :param uut_gateway: Device Gateway
        :param asa_image: asa boot image to be transferred via TFTP,
            e.g. 'asa/asa952-2-smp-k8.bin' for saleen or 'asa/asa952-2-lfbff-k8.SPA' for Kenton
        :param dns_server: DNS server
        :param hostname: hostname to be set
        :param search_domains: search domains delimited by comma,
            defaulted to 'cisco.com'
        :param is_device_kenton: True if device is Kenton, False if device is Saleen
        :param retry_count: download retry count, defaulted to MAX_RETRY_COUNT
        :param power_cycle_flag: True power cylce before baseline, False otherwise
        :param pdu_ip: PDU IP
        :param pdu_port: PDU Port
        :param mgmt_port: name of the management port; if not provided, defaulted to 'Management1/1'
        :param timeout: in seconds; time to wait for fetching the boot image from TFTP server;
                       defaulted to 600s
        :return: None

        """
        self.rommon_tftp_server = rommon_tftp_server
        self.uut_ip = uut_ip
        self.uut_netmask = uut_netmask
        self.uut_gateway = uut_gateway
        self.rommon_image = asa_image
        self.dns_server = dns_server
        self.hostname = hostname
        self.search_domains = search_domains

        if self.sm.current_state is 'rommon_state':
            logger.info('Reboot the device ...')
            self.spawn_id.sendline('reboot')

        if self.sm.current_state is 'boot_state':
            logger.info('Device is in boot_state, reload the device')
            self.spawn_id.sendline('system reload')
            d1 = Dialog([
                ['Are you sure you want to reload the system',
                 'sendline(y)', None, False, False],
            ])
            d1.process(self.spawn_id, timeout=30)

        logger.info('=== Enable configure')
        self.enable_configure()

        logger.info("=== Configure manamgement interface")
        self.configure_mgmt_interface(uut_ip, uut_netmask, uut_gateway, rommon_tftp_server, mgmt_port)

        logger.info("=== Set the new boot image")
        self.set_boot_image(self.rommon_tftp_server, asa_image)

        logger.info('=== Reboot the device ...')
        self.go_to('config_state')
        self.spawn_id.sendline('reload')
        d = Dialog([
            ['Proceed with reload?', 'sendline()', None, True, False],
            ['Rebooting', None, None, False, False],
        ])
        d.process(self.spawn_id, timeout=180)

        logger.info('==== Drop the device to rommon.')
        self.rommon_go_to()
        self.rommon_config(is_device_kenton)

        logger.info('tftpdnld - tftp server: {}, '
                    'rommon image: {} ...'.format(rommon_tftp_server, asa_image))
        self.rommon_boot(timeout=timeout)

        logger.info("=== Configure management interface")
        self.go_to('any')
        self.configure_mgmt_interface(uut_ip, uut_netmask, uut_gateway, rommon_tftp_server, mgmt_port)
        logger.info("=== You have converted successfully the device to Elektra")
 
    def baseline_by_branch_and_version(self, site, branch, version,
                                       uut_ip, uut_netmask, uut_gateway, dns_server, serverIp='', tftpPrefix='', scpPrefix='', docs='', **kwargs):
        """Baseline ftd using branch and version
        
        :param site: e.g. 'ful', 'ast'
        :param branch: branch name, e.g. 'Release', 'Feature'
        :param version: software build-version, e,g, 6.2.3-623
        :param uut_ip: Device IP Address to access TFTP Server
        :param uut_netmask: Device Netmask
        :param uut_gateway: Device Gateway
        :param dns_server: DNS server
        :param \**kwargs:
            :Keyword Arguments, any of below optional parameters:
        :param hostname: hostname to be set; defaulted to 'firepower'
        :param search_domains: search domains delimited by comma,
            defaulted to 'cisco.com'
        :param is_device_kenton: True if device is Kenton, False if device is Saleen
        :param retry_count: download retry count, defaulted to MAX_RETRY_COUNT
        :param power_cycle_flag: True power cylce before baseline, False otherwise
        :param pdu_ip: PDU IP
        :param pdu_port: PDU Port
        :param pdu_user: PDU admn
        :param pdu_pwd: PDU pwd
        :param ntp_server: NTP server delimited by comma, defaulted to None,
            otherwise the value of ntp_server is - e.g. "ntp.esl.cisco.com"
        :param mode: the manager mode (local, remote)
        :param uut_ip6: Device IPv6 Address
        :param uut_prefix: Device IPv6 Prefix
        :param uut_gateway6: Device IPv6 Gateway
        :param manager: FMC to be configured for registration
        :param manager_key: Registration key
        :param manager_nat_id: Registration NAT Id
        :param firewall_mode: the firewall mode (routed, transparent, ngips)
        :param timeout: in seconds; time to wait for fetching the boot image from TFTP server;
                        defaulted to 600s
        :return: None
        """

        is_device_kenton = kwargs.get('is_device_kenton', True)

        if KICK_EXTERNAL:
            server_ip = serverIp
            tftp_prefix = tftpPrefix
            files = docs
        else:
            device_type = 'kenton' if is_device_kenton else 'saleen'
            server_ip, tftp_prefix, scp_prefix, files = prepare_installation_files(site, device_type, branch, version)

        rommon_file_extension = '.lfbff' if is_device_kenton else '.cdisk'
        rommon_file = [file for file in files if file.endswith(rommon_file_extension)]

        files.remove(rommon_file[0])
        pkg_image = "http://{}/{}/{}".format(server_ip, tftp_prefix[len('asa/'):], files[0])
        rommon_image = os.path.join(tftp_prefix, rommon_file[0])

        kwargs['rommon_tftp_server'] = server_ip
        kwargs['pkg_image'] = pkg_image
        kwargs['rommon_image'] = rommon_image
        kwargs['uut_ip'] = uut_ip
        kwargs['uut_netmask'] = uut_netmask
        kwargs['uut_gateway'] = uut_gateway
        kwargs['dns_server'] = dns_server

        self.rommon_to_new_image(**kwargs)



import logging

from unicon.eal.dialogs import Dialog

logger = logging.getLogger(__name__)


class ManagementBootstrap:
    def __init__(self, handle, ipv4_address=None, ipv4_netmask=None, ipv4_gateway=None, ipv6_address=None,
                 ipv6_prefix=None, ipv6_gateway=None, registration_key=None, password=None, firepower_manager_ip=None,
                 firewall_mode=None, fqdn=None, dns_servers=None, search_domains=None, permit_expert_mode=None,
                 management_type=None):
        """
        :param handle: SSH connection handle
        """
        self.handle = handle
        self.ipv4_address = ipv4_address
        self.ipv4_netmask = ipv4_netmask
        self.ipv4_gateway = ipv4_gateway
        self.ipv6_address = ipv6_address
        self.ipv6_prefix = ipv6_prefix
        self.ipv6_gateway = ipv6_gateway
        self.registration_key = registration_key
        self.password = password
        self.firepower_manager_ip = firepower_manager_ip
        self.firewall_mode = firewall_mode
        self.fqdn = fqdn
        self.dns_servers = dns_servers
        self.search_domains = search_domains
        self.permit_expert_mode = permit_expert_mode
        self.management_type = management_type

    def configure(self, slot, app_name):
        logger.info('Configuring App mgmt details in mgmt-bootstrap')
        self.handle.execute('enter mgmt-bootstrap {}'.format(app_name))
        self.configure_ipv4(slot=slot, app_name=app_name, ipv4_address=self.ipv4_address,
                            ipv4_netmask=self.ipv4_netmask, ipv4_gateway=self.ipv4_gateway, scope=False, commit=False)
        self.configure_ipv6(slot=slot, app_name=app_name, ipv6_address=self.ipv6_address,
                            ipv6_prefix=self.ipv6_prefix, ipv6_gateway=self.ipv6_gateway, scope=False, commit=False)
        self.configure_bootstrap_keys_secret(app_name=app_name, registration_key=self.registration_key,
                                             password=self.password, scope=False, commit=False)
        self.configure_bootstrap_keys(app_name=app_name, firepower_manager_ip=self.firepower_manager_ip,
                                      firewall_mode=self.firewall_mode, fqdn=self.fqdn, dns_servers=self.dns_servers,
                                      search_domains=self.search_domains, scope=False, commit=False,
                                      permit_expert_mode=self.permit_expert_mode, management_type=self.management_type)
        logger.info('Configured App mgmt details in mgmt-bootstrap')

    def configure_ipv4(self, slot, app_name, ipv4_address=None, ipv4_netmask=None, ipv4_gateway=None, scope=True,
                       commit=True):
        logger.info('Configuring IPv4 details in mgmt-bootstrap')
        if ipv4_address or ipv4_netmask or ipv4_gateway:
            if scope:
                self.handle.execute('enter mgmt-bootstrap {}'.format(app_name))
            self.handle.execute('enter {} {} firepower'.format('ipv4', slot))
            if ipv4_gateway:
                self.handle.execute('set gateway {}'.format(ipv4_gateway))
            if ipv4_address and ipv4_netmask:
                self.handle.execute('set ip {} {} {}'.format(ipv4_address, 'mask', ipv4_netmask))
            elif ipv4_address or ipv4_netmask:
                raise Exception("You need to provide both ipv4_address and ipv4_netmask")
            if commit:
                output = self.handle.execute('commit-buffer')
                if 'Error' in output:
                    logger.info("Failed while configuring IPv4 with error: {}".format(output))
                    raise Exception("Failed while configuring IPv4")
            self.handle.execute('exit')
        logger.info('Completed IPv4 configuration in mgmt-bootstrap')

    def configure_ipv6(self, slot, app_name, ipv6_address=None, ipv6_prefix=None, ipv6_gateway=None, scope=True,
                       commit=True):
        logger.info('Configuring IPv6 details in mgmt-bootstrap')
        if ipv6_address or ipv6_prefix or ipv6_gateway:
            if scope:
                self.handle.execute('enter mgmt-bootstrap {}'.format(app_name))
            self.handle.execute('enter {} {} firepower'.format('ipv6', slot))
            if ipv6_gateway:
                self.handle.execute('set gateway {}'.format(ipv6_gateway))
            if ipv6_address and ipv6_prefix:
                self.handle.execute('set ip {} {} {}'.format(ipv6_address, 'prefix-length', ipv6_prefix))
            elif ipv6_address or ipv6_prefix:
                raise Exception("You need to provide both ipv6_address and ipv6_prefix")
            if commit:
                output = self.handle.execute('commit-buffer')
                if 'Error' in output:
                    logger.info("Failed while configuring IPv6 with error: {}".format(output))
                    raise Exception("Failed while configuring IPv6")
            self.handle.execute('exit')
        logger.info('Completed IPv6 configuration in mgmt-bootstrap')

    def configure_bootstrap_keys_secret(self, app_name, registration_key=None, password=None, scope=True, commit=True):
        logger.info('Configuring bootstrap key secret in mgmt-bootstrap')
        if registration_key or password:
            if scope:
                self.handle.execute('enter mgmt-bootstrap {}'.format(app_name))
            if registration_key:
                self._enter_secret_value("REGISTRATION_KEY", registration_key)
            if password:
                self._enter_secret_value("PASSWORD", password)
            if commit:
                output = self.handle.execute('commit-buffer')
                if 'Error' in output:
                    logger.info("Failed while configuring bootstrap key with error: {}".format(output))
                    raise Exception("Failed while configuring bootstrap key")
        logger.info('Completed bootstrap key secret configuration in mgmt-bootstrap')

    def _enter_secret_value(self, type, value):
        self.handle.execute('enter bootstrap-key-secret {}'.format(type))
        configuration_dialog = Dialog([
            ['Enter a value', 'sendline({})'.format(value), None, True, True],
            ['Confirm the value:', 'sendline({})'.format(value), None, False, False],
        ])
        target_state = self.handle.sm.current_state
        self.handle.run_cmd_dialog('set value', configuration_dialog, target_state=target_state)
        self.handle.execute('exit')

    def configure_bootstrap_keys(self, app_name, firepower_manager_ip=None,
                                 firewall_mode=None, fqdn=None, dns_servers=None, search_domains=None, scope=True,
                                 commit=True, permit_expert_mode=None, management_type=None):
        logger.info('Configuring bootstrap key in mgmt-bootstrap')
        if any([firepower_manager_ip, firewall_mode, fqdn, dns_servers, search_domains, permit_expert_mode,
                management_type]):
            if scope:
                self.handle.execute('enter mgmt-bootstrap {}'.format(app_name))
            self._enter_value("FIREPOWER_MANAGER_IP", firepower_manager_ip)
            self._enter_value("FIREWALL_MODE", firewall_mode)
            self._enter_value("FQDN", fqdn)
            self._enter_value("DNS_SERVERS", dns_servers)
            self._enter_value("SEARCH_DOMAINS", search_domains)
            self._enter_value("PERMIT_EXPERT_MODE", permit_expert_mode)
            self._enter_value("MANAGEMENT_TYPE", management_type)
            if commit:
                output = self.handle.execute('commit-buffer')
                if 'Error' in output:
                    logger.info("Failed while configuring bootstrap key with error: {}".format(output))
                    raise Exception("Failed while configuring bootstrap key")
        logger.info('Completed bootstrap key configuration in mgmt-bootstrap')

    def _enter_value(self, type, value):
        if value is not None:
            self.handle.execute('enter bootstrap-key {}'.format(type))
            self.handle.execute('set value {}'.format(value))
            self.handle.execute('up')


class ManagementClusterBootstrap(ManagementBootstrap):

    def __init__(self, handle, ip_config=None, registration_key=None, password=None,
                 firepower_manager_ip=None, firewall_mode=None, fqdn=None, dns_servers=None, search_domains=None,
                 permit_expert_mode=None, management_type=None):
        """
        :param handle: SSH connection handle
        """
        self.handle = handle
        self.ip_config = ip_config
        self.registration_key = registration_key
        self.password = password
        self.firepower_manager_ip = firepower_manager_ip
        self.firewall_mode = firewall_mode
        self.fqdn = fqdn
        self.dns_servers = dns_servers
        self.search_domains = search_domains
        self.permit_expert_mode = permit_expert_mode
        self.management_type = management_type

    def configure(self, slot, app_name):
        logger.info('Configuring App mgmt details in mgmt-bootstrap for Intra chassis cluster')
        self.handle.execute('enter mgmt-bootstrap {}'.format(app_name))
        if ',' in slot:
            slot_list = slot.strip("\"").split(',')
            for app_slot in slot_list:
                if app_slot in self.ip_config:
                    if 'ipv4' in self.ip_config[app_slot]:
                        self.configure_ipv4(slot=app_slot, app_name=app_name,
                                            ipv4_address=self.ip_config[app_slot]['ipv4'].get("ip"),
                                            ipv4_netmask=self.ip_config[app_slot]['ipv4'].get("netmask"),
                                            ipv4_gateway=self.ip_config[app_slot]['ipv4'].get("gateway"),
                                            scope=False, commit=False)
                    if 'ipv6' in self.ip_config[app_slot]:
                        self.configure_ipv6(slot=app_slot, app_name=app_name,
                                            ipv6_address=self.ip_config[app_slot]['ipv6'].get("ip"),
                                            ipv6_prefix=self.ip_config[app_slot]['ipv6'].get("prefix"),
                                            ipv6_gateway=self.ip_config[app_slot]['ipv6'].get("gateway"),
                                            scope=False, commit=False)

        self.configure_bootstrap_keys_secret(app_name=app_name, registration_key=self.registration_key,
                                             password=self.password, scope=False, commit=False)
        self.configure_bootstrap_keys(app_name=app_name, firepower_manager_ip=self.firepower_manager_ip,
                                      firewall_mode=self.firewall_mode, fqdn=self.fqdn, dns_servers=self.dns_servers,
                                      search_domains=self.search_domains, scope=False, commit=False,
                                      permit_expert_mode=self.permit_expert_mode, management_type=self.management_type)
        logger.info('Configured App mgmt details in mgmt-bootstrap')



import logging

from unicon.eal.dialogs import Dialog

logger = logging.getLogger(__name__)


class ClusterBootstrap:
    def __init__(self, handle, chassis_id=None, ccl_network=None, cluster_key=None, cluster_name=None,
                 site_id=None):
        """
        :param handle: SSH connection handle
        """
        self.handle = handle
        self.chassis_id = chassis_id
        self.ccl_network = ccl_network
        self.cluster_key = cluster_key
        self.cluster_name = cluster_name
        self.site_id = site_id


    def configure(self):
        logger.info('Configuring cluster mgmt details in cluster-bootstrap')
        self.handle.execute('enter cluster-bootstrap')
        if self.chassis_id is not None:
            self.handle.execute("set chassis-id {}".format(self.chassis_id))
        if self.ccl_network is not None:
            self.handle.execute("set cluster-control-link network {}".format(self.ccl_network))
        self.handle.execute("""set ipv4 gateway 0.0.0.0
                               set ipv4 pool 0.0.0.0 0.0.0.0
                               set ipv6 gateway ::
                               set ipv6 pool :: ::
                               set virtual ipv4 0.0.0.0 mask 0.0.0.0
                               set virtual ipv6 :: prefix-length ""
                               set mode spanned-etherchannel""")
        if self.cluster_name is not None:
            self.handle.execute("set name {}".format(self.cluster_name))
        if self.site_id is not None:
            self.handle.execute("set site-id {}".format(self.site_id))
        if self.cluster_key is not None:
            self._enter_key_value(self.cluster_key)
        self.handle.execute('exit')
        logger.info('Configured cluster-bootstrap details')



    def _enter_key_value(self,value):
        configuration_dialog = Dialog([
            ['Enter a key', 'sendline({})'.format(value), None, True, True],
            ['Confirm the key:', 'sendline({})'.format(value), None, False, False],
        ])
        target_state = self.handle.sm.current_state
        self.handle.run_cmd_dialog('set key', configuration_dialog, target_state=target_state)



import logging
import re

from kick.device2.fxos.resource import ResourceProfile
from kick.device2.fxos.utils import scope

logger = logging.getLogger(__name__)


class ResourceProfileUtils:
    def __init__(self, handle):
        """
        Constructor of App Instance
        :param handle: SSH connection handle
        """
        self.handle = handle

    def get_all(self):
        """
        Get all resource profiles
        :return:
        """
        scope(self.handle, 'ssa')
        profiles = []
        output = self.handle.execute('show resource-profile').split('\r\n')
        is_profile = False
        for i in output:
            if is_profile and re.match("^([\w\-]+)", i):
                match = re.match("^([\w\-]+)", i)
                profiles.append(match.group(1))
            if '---' in i:
                is_profile = True
        return profiles

    def delete_all(self):
        """
        Delete all resource profiles
        :return:
        """
        all_profiles = self.get_all()
        for profile in all_profiles:
            resource_profile = ResourceProfile(self.handle, profile)
            resource_profile.delete()

    def check_profile_present(self, name):
        """
        Check if resource profile exists
        :return:
        """
        all_profiles = self.get_all()
        for profile in all_profiles:
            if profile == name:
                return True
        return False



import logging
import re

from kick.device2.fxos.interfaces.physical_interface import PhysicalInterface
from kick.device2.fxos.interfaces.portchannel_interface import PortChannelInterface
from kick.device2.fxos.interfaces.portchannel_subinterface import PortChannelSubInterface
from kick.device2.fxos.interfaces.subinterface import SubInterface
from kick.device2.fxos.utils import scope

logger = logging.getLogger(__name__)


class InterfaceHandler:

    def __init__(self, handle):
        """
        Create a dictionary with all the required info of the interface that can be used by functions
        :param handle: ssh connection handle
        """
        self.handle = handle

    def get_all_port_channel(self):
        """
        get all port channel interfaces
        """
        logger.info('Get all Port Channel Interface')
        scope(self.handle, 'eth-uplink/fabric')
        output = self.handle.execute('show port-channel').split('\r\n')
        portchannel = []
        for i in output:
            if re.match("^\s+\d+\s+(Port-channel\d+)", i):
                match = re.match("^\s+\d+\s+(Port-channel\d+)", i)
                portchannel.append(match.group(1))
        return portchannel

    def delete_all_port_channel(self):
        all_port_channel = self.get_all_port_channel()
        for name in all_port_channel:
            if name != 'Port-channel48':
                interface_obj = PortChannelInterface(self.handle, name)
                interface_obj.delete()

    def instantiate_interface_based_on_name(self, name):
        """
        Instantiates the interface of the correct type based on its name
        :param name: interface name
        """
        if re.match("^\w+\/\d+$", name):
            return PhysicalInterface(self.handle, name)
        elif re.match("^\w+\/\d+\.\d+$", name):
            match = re.match("^(\w+\/\d+)\.(\d+)$", name)
            return SubInterface(self.handle, match.group(1), match.group(2))
        elif re.match("^Port-channel\d+$", name):
            return PortChannelInterface(self.handle, name)
        elif re.match("^(Port-channel\d+)\.(\d+)$", name):
            match = re.match("^(Port-channel\d+)\.(\d+)$", name)
            return PortChannelSubInterface(self.handle, match.group(1), match.group(2))
        else:
            logger.error("Interface Name {} is not valid".format(name))
            raise Exception("Name is not valid")

    def get_all_physical_interfaces(self):
        """
        get all physical interfaces
        """
        logger.info('Get all Physical Interface')
        scope(self.handle, 'eth-uplink/fabric')
        output = self.handle.execute('show interface').split('\r\n')
        interfaces = []
        for i in output:
            if re.match("^\s+(\w+\/\d+)", i):
                match = re.match("^\s+(\w+\/\d+)", i)
                interfaces.append(match.group(1))
        return interfaces

    def reset_all_physical_interfaces(self):
        """
        get all physical interfaces
        """
        interfaces = self.get_all_physical_interfaces()
        for name in interfaces:
            interface_obj = PhysicalInterface(self.handle, name)
            interface_obj.reset()
            subinterfaces = interface_obj.get_subinterfaces()
            for sub_id in subinterfaces:
                subinterface_obj = SubInterface(self.handle, name, sub_id)
                subinterface_obj.delete()



from unicon.statemachine import State, Path, StateMachine
from unicon.eal.dialogs import Dialog
from .patterns import Ftd5500xPatterns
from .statements import Ftd5500xStatements
from .dialogs import Ftd5500xDialog


class Ftd5500xStatemachine(StateMachine):
    def __init__(self, patterns):
        """Initializer of FtdStateMachine."""
        self.patterns = patterns
        self.dialogs = Ftd5500xDialog(patterns)
        self.statements = Ftd5500xStatements(patterns)

        super().__init__()

    def create(self):
        # Create States
        prelogin_state = State('prelogin_state', self.patterns.prompt.prelogin_prompt)
        fireos_state = State('fireos_state', self.patterns.prompt.fireos_prompt)
        expert_state = State('expert_state', self.patterns.prompt.expert_prompt)
        sudo_state = State('sudo_state', self.patterns.prompt.sudo_prompt)
        enable_state = State('enable_state', self.patterns.prompt.enable_prompt)
        ciscoasa_state = State('ciscoasa_state', self.patterns.prompt.ciscoasa_prompt)
        rommon_state = State('rommon_state', self.patterns.prompt.rommon_prompt)
        boot_state = State('boot_state', self.patterns.prompt.firepower_boot)
        disable_state = State('disable_state', self.patterns.prompt.disable_prompt)
        config_state = State('config_state', self.patterns.prompt.config_prompt)
        system_state = State('system_state', self.patterns.prompt.system_prompt)

        # Add states
        self.add_state(prelogin_state)
        self.add_state(fireos_state)
        self.add_state(expert_state)
        self.add_state(sudo_state)
        self.add_state(enable_state)
        self.add_state(ciscoasa_state)
        self.add_state(rommon_state)
        self.add_state(boot_state)
        self.add_state(disable_state)
        self.add_state(config_state)
        self.add_state(system_state)

        # Create paths
        prelogin_to_fireos_path = Path(prelogin_state, fireos_state, 'admin',
                                       Dialog([self.statements.login_password_statement]))
        fireos_to_prelogin_path = Path(fireos_state, prelogin_state, 'exit', None)

        fireos_to_expert_path = Path(fireos_state, expert_state, 'expert', None)
        expert_to_fireos_path = Path(expert_state, fireos_state, 'exit', None)
        expert_to_sudo_path = Path(expert_state, sudo_state, 'sudo su -',
                                   Dialog([self.statements.sudo_password_statement]))
        sudo_to_expert_path = Path(sudo_state, expert_state, 'exit', None)

        ciscoasa_state_to_sudo_state = Path(ciscoasa_state, sudo_state, 'sudo su -',
                                            Dialog([self.statements.sudo_password_statement]))
        fireos_to_ciscoasa_state = Path(fireos_state, ciscoasa_state, 'expert', None)

        # lina cli paths
        fireos_to_disable_state = Path(fireos_state, disable_state,
                                       'system support diagnostic-cli',
                                       None)
        disable_to_enable_state = Path(disable_state, enable_state, 'en',
                                       self.dialogs.disable_to_enable)
        enable_to_disable_state = Path(enable_state, disable_state, "disable", None)

        disable_to_fireos_path = Path(disable_state, fireos_state, '\001'+'d'+'exit',
                                      None)

        enable_to_config_path = Path(enable_state, config_state, 'conf t', None)
        config_to_enable_path = Path(config_state, enable_state, 'end', None)

        # system paths
        system_to_expert_path = Path(system_state, expert_state, 'expert', None)
        expert_to_system_path = Path(expert_state, system_state, 'exit', None)

        # Add paths
        self.add_path(prelogin_to_fireos_path)
        self.add_path(fireos_to_prelogin_path)
        self.add_path(fireos_to_expert_path)
        self.add_path(expert_to_fireos_path)
        self.add_path(expert_to_sudo_path)
        self.add_path(sudo_to_expert_path)
        self.add_path(disable_to_fireos_path)
        self.add_path(ciscoasa_state_to_sudo_state)
        self.add_path(fireos_to_ciscoasa_state)
        self.add_path(fireos_to_disable_state)
        self.add_path(disable_to_enable_state)
        self.add_path(enable_to_config_path)
        self.add_path(config_to_enable_path)
        self.add_path(system_to_expert_path)
        self.add_path(expert_to_system_path)
        self.add_path(enable_to_disable_state)

        # Add a default statement:
        self.add_default_statements(self.statements.enable_password_statement)



"""Copyright (c) 2017 Cisco Systems, Inc.
Name:
    asa_cluster.py
Usage:
    Submodule for Legacy ASA Diameter feature.
Author:
    raywa
"""

import re
from .asa_config import AsaConfig

class AsaDiameterConfig(AsaConfig):
    """ASA Config for Diameter inherited from AsaConfig
    """
    def __init__(self, **kwargs):
        """Initializer of AsaDiameterConfig
        """
        super().__init__(**kwargs)

    def get_service_policy_stats(self, ctx=None):
        """Show Diameter service policy statistics
        Example output:
            gtp-spykerd(config)# show service-policy | include diameter
            Inspect: diameter diametermap, packet 4073569, lock fail 0, drop 0, reset-drop 0,
            5-min-pkt-rate 0 pkts/sec, v6-fail-close 0 sctp-drop-override 0

        :param ctx: context name for multi mode only
        :return: dict of stats parameters

        """
        return_value = {}
        if self.topo.mode == 'standalone':
            output = self.execute('show service-policy | include diameter', ctx).strip().split(',')
            return_value['policy_map'] = output[0].split()[-1]
            for param in output[1:]:
                param = param.strip()
                while True:
                    # pylint: disable=anomalous-backslash-in-string
                    match = re.search(' (\d+)', param)
                    if match:
                        return_value[param[:match.start()].strip()] = int(match.group(1))
                        param = param[match.end():]
                    else:
                        break
        elif self.topo.mode == 'cluster':
            pass
            # not supported now
        return return_value
        


"""ASA Constants."""

from enum import Enum

class AsaSmStates(Enum):
    ENABLE_STATE = 'enable_state'
    DISABLE_STATE = 'disable_state'
    CONFIG_STATE = 'config_state'

class AsaConfigConstants:
    FIREWALL_MODE = dict(rfw='Router', tfw='Transparent')
    CONTEXT_MODE = dict(sfm='single', mfm='multiple')
    DELETE_NOCONFIRM = '/noconfirm'
    DISK0 = 'disk0'
    STATE = dict(disabled=False, disable=False, enabled=True, enable=True, false=False, true=True)
    STATE['False'] = False
    STATE['True'] = True

class ASAClusterStates(Enum):
    MASTER = 'MASTER'
    SLAVE = 'SLAVE'
    DISABLED = 'DISABLED'

class ASAHAStates(Enum):
    PRIMARY = 'Primary'
    SECONDARY = 'Secondary'
    ACTIVE = 'Active'
    STANDBY = 'Standby Ready'

class ASAHALinkNames(Enum):
    DEFAULT_FAILOVER_LINK = 'folink'
    DEFAULT_STATE_LINK = 'statelink'




from ...m4.actions import M4, M4Line

try:
    from kick.graphite.graphite import publish_kick_metric
except ImportError:
    from kick.metrics.metrics import publish_kick_metric

KICK_EXTERNAL = False

try:
    from kick.file_servers.file_servers import *
except ImportError:
    KICK_EXTERNAL = True
    pass

try:
    from kick.kick_constants import KickConsts
except ImportError:
    from kick.miscellaneous.credentials import KickConsts


import logging

logger = logging.getLogger(__name__)


class M5(M4):
    def __init__(self, hostname='firepower', login_username='admin',
                 login_password=KickConsts.DEFAULT_PASSWORD,
                 sudo_password=KickConsts.DEFAULT_PASSWORD,
                 cimc_hostname='', *args, **kwargs):
        """Constructor for FMC M5.

        :param hostname: fmc hostname or fqdn e.g. FS2000-01 or FS2000-01.cisco.com
        :param login_username: user credentials
        :param login_password: device login password with user name 'admin'
        :param sudo_password: device sudo password for 'root'

        """

        super().__init__(hostname, login_username,
                         login_password, sudo_password, cimc_hostname)
        publish_kick_metric('device.m5.init', 1)

        self.line_class = M5Line


class M5Line(M4Line):
    def __init__(self, spawn_id, sm, type, timeout=None):
        super().__init__(spawn_id, sm, type, timeout)
        self.metric = 'm5'

    def baseline_fmc_m5(self, iso_map_name, http_link, iso_file_name,
                        mgmt_ip, mgmt_netmask, mgmt_gateway, timeout=None,
                        mgmt_ip6=None, mgmt_prefix=None, mgmt_gateway6=None,
                        change_password=True):
        """Baseline M5 through CIMC connection.

        :param iso_map_name: map name for command map-www, e.g. 'myiso'
        :param http_link: http link for the share, e.g. 'http://10.83.65.25'
        :param iso_file_name: iso filename under http_link,
               e.g. '/cache/Development/6.3.0-10558/iso/Sourcefire_Defense_Center-6.3.0-10558-Autotest.iso
        :param mgmt_ip: management interface ip address
        :param mgmt_netmask: management interface netmask
        :param mgmt_gateway: management interface gateway
        :param timeout: in seconds; time to wait for device to boot with the new image;
                        if not provided, default baseline time is 7200s
        :param mgmt_ip6: management interface IPv6 Address
        :param mgmt_prefix: management interface IPv6 Prefix
        :param mgmt_gateway6: management interface IPv6 Gateway
        :param change_password: Flag to change the password after baseline
        :return: None

        """

        super().baseline_fmc_m4(iso_map_name=iso_map_name,
                                http_link=http_link, iso_file_name=iso_file_name,
                                mgmt_ip=mgmt_ip, mgmt_netmask=mgmt_netmask,
                                mgmt_gateway=mgmt_gateway, timeout=timeout,
                                mgmt_ip6=mgmt_ip6, mgmt_prefix=mgmt_prefix,
                                mgmt_gateway6=mgmt_gateway6,
                                change_password=change_password)

        # workaround to enable admin user
        # to be removed when issue is solved
        self._enable_admin_user()

    def _enable_admin_user(self):
        """Temporary workaround to access CIMC with admin user
        :return:
        """
        self.go_to("mio_state")
        self.execute("top")
        self.execute("scope user 1")
        self.execute("set enabled yes")
        self.execute("commit")

    def baseline_by_branch_and_version(self, site, branch, version, uut_ip, uut_netmask,
                                       uut_gateway, iso_file_type='Restore', timeout=None,
                                       mgmt_ip6=None, mgmt_prefix=None, mgmt_gateway6=None,
                                       change_password=True, cimc=True, serverIp='', tftpPrefix='', scpPrefix='', docs='', **kwargs):
        """Baseline of FMC M5 through CIMC connection by branch and version using PXE servers.
        Looks for needed files on devit-engfs, copies them to the local kick server
        and use them to baseline the device.

        :param site: e.g. 'ful', 'ast', 'bgl'
        :param branch: branch name, e.g. 'Release', 'Feature'
        :param version: software build-version, e,g, 6.2.3-623
        :param uut_ip: ip address of management interface
        :param uut_netmask: netmask of management interface
        :param uut_gateway: gateway ip address of mgmt interface
        :param iso_file_type: 'Autotest' or 'Restore'; defaulted to 'Restore'
        :param timeout: in seconds; time to wait for device to boot with the new image;
                        if not provided, default baseline time is 7200s
        :param mgmt_ip6: management interface IPv6 Address
        :param mgmt_prefix: management interface IPv6 Prefix
        :param mgmt_gateway6: management interface IPv6 Gateway
        :param change_password: Flag to change the password after baseline
        :param cimc: Flag to baseline using CIMC or serial
        **kwargs:
        :param power_cycle_flag: True power cycle before baseline, False otherwise
        :param pdu_ip: string of IP addresses of the PDU's
        :param pdu_port: string of power port on the PDU's
        :param pdu_user: usernames for power bar servers
        :param pdu_pwd: passwords for power bar servers
        :param mgmt_ip6: management interface IPv6 Address
        :param mgmt_prefix: management interface IPv6 Prefix
        :param mgmt_gateway6: management interface IPv6 Gateway
        :param change_password: Flag to change the password after baseline
        :param ipv4_mode: IPv4 configuration mode - 'static' or 'dhcp';
                         default set to 'static'
        :param ipv6_mode: IPv6 configuration mode - 'dhcp', 'router' or 'manual'
                          default set to 'manual'
        :param dns_servers: a comma-separated string of DNS servers
        :param search_domains: a comma-separated string of search domains
                                default set to 'cisco.com'
        :param ntp_servers: a comma-separated string of NTP servers
        """

        if KICK_EXTERNAL:
            server_ip = serverIp
            tftp_prefix = tftpPrefix
            scp_prefix = scpPrefix
            files = docs
        else:
            server_ip, tftp_prefix, scp_prefix, files = prepare_installation_files(site, 'm5', branch, version,
                                                                               iso=iso_file_type)
        if not files[0]:
            raise Exception('M5 iso file not found on server')

        kwargs['mgmt_ip'] = uut_ip
        kwargs['mgmt_netmask'] = uut_netmask
        kwargs['mgmt_gateway'] = uut_gateway
        kwargs['mgmt_ip6'] = mgmt_ip6
        kwargs['mgmt_prefix'] = mgmt_prefix
        kwargs['mgmt_gateway6'] = mgmt_gateway6
        kwargs['timeout'] = timeout
        kwargs['change_password'] = change_password
        if cimc:
            kwargs['iso_map_name'] = 'myiso'
            kwargs['http_link'] = 'http://{}'.format(server_ip)
            kwargs['iso_file_name'] = '{}/{}'.format(tftp_prefix[len('asa'):], files[0])
            self.baseline_fmc_m5(**kwargs)
        else:
            kwargs['iso_url'] = 'http://{}/{}/{}'.format(server_ip, tftp_prefix[len('asa'):], files[0])
            self.baseline_using_serial(**kwargs)

    def baseline_using_serial(self, iso_url, mgmt_ip, mgmt_netmask, mgmt_gateway, mgmt_intf="eth0", timeout=None,
                              power_cycle_flag=False, pdu_ip='', pdu_port='',
                              pdu_user='admn', pdu_pwd='admn',
                              mgmt_ip6=None, mgmt_prefix=None, mgmt_gateway6=None,
                              change_password=True, ipv4_mode='static', ipv6_mode='',
                              dns_servers='', search_domains='', ntp_servers=''):
        """Baseline FMC M5 device through its physical serial port connection

        :param iso_url: http url of iso image
                http://10.83.65.25/cache/Development/6.3.0-10581/iso/Sourcefire_Defense_Center-6.3.0-10581-Restore.iso
        :param mgmt_ip: management interface ip address
        :param mgmt_netmask: management interface netmask
        :param mgmt_intf: management interface gateway
        :param mgmt_gateway: management interface gateway
        :param timeout: in seconds; time to wait for device to boot with the new image;
                        if not provided, default baseline time is 7200s
        :param power_cycle_flag: True power cylce before baseline, False otherwise
        :param pdu_ip: string of IP addresses of the PDU's
        :param pdu_port: string of power port on the PDU's
        :param pdu_user: usernames for power bar servers
        :param pdu_pwd: passwords for power bar servers
        :param mgmt_ip6: management interface IPv6 Address
        :param mgmt_prefix: management interface IPv6 Prefix
        :param mgmt_gateway6: management interface IPv6 Gateway
        :param change_password: Flag to change the password after baseline
        :param ipv4_mode: IPv4 configuration mode - 'static' or 'dhcp'
        :param ipv6_mode: IPv6 configuration mode - 'dhcp', 'router' or 'manual'
        :param dns_servers: a comma-separated string of DNS servers
        :param search_domains: a comma-separated string of search domains
        :param ntp_servers: a comma-separated string of NTP servers
        :return:
        """

        super().baseline_using_serial(iso_url=iso_url, mgmt_ip=mgmt_ip, mgmt_netmask=mgmt_netmask,
                                      mgmt_gateway=mgmt_gateway, mgmt_intf=mgmt_intf, timeout=timeout,
                                      power_cycle_flag=power_cycle_flag, pdu_ip=pdu_ip, pdu_port=pdu_port,
                                      pdu_user=pdu_user, pdu_pwd=pdu_pwd,
                                      mgmt_ip6=mgmt_ip6, mgmt_prefix=mgmt_prefix, mgmt_gateway6=mgmt_gateway6,
                                      change_password=change_password, ipv4_mode=ipv4_mode, ipv6_mode=ipv6_mode,
                                      dns_servers=dns_servers, search_domains=search_domains, ntp_servers=ntp_servers)



import logging

import time

from kick.device2.fxos.app_handler import AppInstanceUtils
from kick.device2.fxos.utils import create_dict, scope, poll

logger = logging.getLogger(__name__)


class AppInstance:
    slot_handler = None
    mi_supported = False

    def __init__(self, handle, app_name, slot, identifier=""):
        """
        Constructor of App Instance
        :param handle: SSH connection handle
        :param app_name: application instance name; e.g. 'ftd'
        :param slot: slot id; e.g. '1', '2'
        :param identifier: identifier or logical device name; e.g. 'ftd1'
        """
        self.handle = handle
        self.app_name = app_name
        self.slot = str(slot)
        self.identifier = identifier
        if self.slot == "":
            raise ValueError('Slot is mandatory and was sent as empty string')
        self.is_mi_supported()

    def get_slot_handler(self):
        if not self.slot_handler:
            self.slot_handler = AppInstanceUtils(handle=self.handle, slot=self.slot)
        return self.slot_handler

    def create(self, deploy_type="native", version=None, resource_profile=None, hw_crypto_state=None):
        """
        Create app instance
        :param deploy_type: deploy type; e.g. 'native', 'container'
        :param version: version; e.g. '6.3.0.10396'
        :param resource_profile: the resource profile name
        :return:
        """
        self.handle.execute("top")
        self.handle.execute("discard-buffer")

        logger.info("Enable license for CSP")
        self.handle.execute("top")
        self.handle.execute("scope ssa")
        self.handle.execute("scope app ftd {}".format(version))
        self.handle.execute("accept-license-agreement", timeout=60)
        output = self.handle.execute("commit-buffer")
        if 'Error' in output:
            raise Exception("Failed to enable FTD license")
        output = self.handle.execute("show detail", timeout=60)
        if 'Has License Agreement: Yes' not in output or 'License Agreement has been Accepted: Yes' not in output:
            raise Exception("Failed to enable FTD license")

        self.handle.execute("scope ssa")
        self.handle.execute("scope slot " + self.slot)

        if self.mi_supported:
            if self.identifier == "":
                raise Exception('Identifier not provided, which is mandatory in the new MI builds even for native apps')

            output = self.handle.execute("create app-instance " + self.app_name + " " + self.identifier)
            if 'Error' in output:
                logger.info("Failed while creating app instance with error: {}".format(output))
                raise Exception("Failed while creating app instance")
            self.handle.execute("set deploy-type " + deploy_type)
        else:
            self.handle.execute("create app-instance " + self.app_name)

        if version:
            output = self.handle.execute('set startup-version ' + version)
            if 'Error' in output or 'Warning' in output:
                logger.info(
                    "Failed while adding startup-version to application instance with error: {}".format(
                        output))
                raise Exception("Failed while creating app instance")
        if resource_profile:
            output = self.handle.execute('set resource-profile-name ' + resource_profile)
            if 'Error' in output or 'Warning' in output:
                logger.info(
                    "Failed while adding resource profile to application instance with error: {}".format(output))
                raise Exception("Failed while creating app instance")
        if hw_crypto_state:
            output = self.handle.execute('create hw-crypto')
            if 'Error' in output or 'Warning' in output:
                logger.info(
                    "Failed while setting hw crypto to application instance with error: {}".format(output))
                raise Exception("Failed while creating app instance")
            self.handle.execute("set admin-state "+ hw_crypto_state)
            self.handle.execute('exit')
        output = self.handle.execute("commit-buffer")
        if 'Error' in output:
            logger.info("Failed while creating app instance with error: {}".format(output))
            raise Exception("Failed while creating app instance")
        self.handle.execute("discard-buffer")

    def disable(self):
        """
        Disable Application Instance
        :return:
        """
        if self.get_app_instance_detail():
            scope(self.handle, 'app-instance {} {}'.format(self.app_name, self.identifier))
            self.handle.execute('disable')
            output = self.handle.execute("commit")
            if 'Error' in output:
                logger.info("Failed while trying to {} instance with error: {}".format('disable', output))
                raise Exception("Failed while enabling or disabling the app instance")
            self.handle.execute("discard-buffer")
            assert poll(self.handle, command='sh detail',
                        expect='Admin State: Disabled +Oper State: Offline',
                        max_retry=50,
                        interval=15), "App Instance Not Online"
        else:
            raise Exception("The app is not present. Cannot disable it!")

    def enable(self):
        """
        Enable Application Instance
        :return:
        """
        if self.get_app_instance_detail():
            scope(self.handle, 'app-instance {} {}'.format(self.app_name, self.identifier))
            self.handle.execute('enable')
            self.handle.execute("commit")
            self.handle.execute("discard-buffer")
            step = 0
            while step < 50:
                logger.info('Checking if enable - {} try'.format(step))
                if self.check_status(move_to_scope=False):
                    return True
                time.sleep(15)
                step += 1
            raise Exception("The app is not enabled!")
        else:
            raise Exception("The app is not present. Cannot enable it!")

    def check_status(self, move_to_scope=True):
        """
        Check status
        :return:
        """
        if move_to_scope:
            scope(self.handle, 'ssa/slot {}'.format(self.slot))
            scope(self.handle, 'app-instance {} {}'.format(self.app_name, self.identifier))
        output = self.handle.execute('sh detail')
        if 'Admin State: Enabled' in output and 'Oper State: Online' in output:
            return True
        if 'Oper State: Start Failed' in output or 'Oper State: Not Available' in output:
            raise Exception('Failed to enable app!')
        return False

    def check_status_in_loop(self, step_no=50, step_wait=30, move_to_scope=True):
        """
        Check status
        :return:
        """
        if self.get_app_instance_detail():
            if move_to_scope:
                scope(self.handle, 'ssa/slot {}'.format(self.slot))
                scope(self.handle, 'app-instance {} {}'.format(self.app_name, self.identifier))
            step = 0
            while step < step_no:
                logger.info('Checking if app is enabled and online - {} try'.format(step))
                if self.check_status(move_to_scope=False):
                    return True
                time.sleep(step_wait)
                step += 1
            raise Exception("The app is not in the correct state!")
        else:
            raise Exception("The app is not present!")

    def delete(self):
        """
        Delete Application Instance
        :return:
        """
        self.handle.execute("top")
        self.handle.execute("discard-buffer")
        self.handle.execute("scope ssa")
        self.handle.execute("scope slot " + self.slot)

        if self.mi_supported:
            if self.identifier == "":
                raise Exception("Identifier is mandatory for MI supported")
            output = self.handle.execute("delete app-instance " + self.app_name + " " + self.identifier)
        else:
            output = self.handle.execute("delete app-instance " + self.app_name)
        if 'Error' in output:
            logger.info("Failed while trying to delete app instance {} with error: {}".format(self.app_name, output))
            raise Exception("Failed while trying to delete app instance")

        self.handle.execute("commit-buffer")
        self.handle.execute("discard-buffer")

    def get_app_instance_detail(self):
        """
        Verify if app instance is present and get app details
        :return: None or Dictionary structure
        """
        self.get_slot_handler()
        for a in self.slot_handler.get_app_instances_in_slot():
            if a['App Name'] == self.app_name and a['Identifier'] == self.identifier:
                return a
        return None

    def get_app_instance_resources(self):
        """
        Gets resources allocated to app-instance
        :param slot: slot id
        :param app: app name
        :param identifier: identifier or logical device name
        :return: resources
        """
        if self.get_app_instance_detail():
            scope(self.handle, 'app-instance {} {}'.format(self.app_name, self.identifier))
            resources = create_dict(self.handle.execute('show resource detail'))
            resources['cpu'] = resources.pop('Allocated Core NR')
            resources['ram'] = resources.pop('Allocated RAM (MB)')
            resources['data_disk'] = resources.pop('Allocated Data Disk (MB)')
            resources['binary_disk'] = resources.pop('Allocated Binary Disk (MB)')
            resources['secondary_disk'] = resources.pop('Allocated Secondary Disk (MB)')
            return resources
        else:
            raise Exception("The app is not present. Cannot get resource profile information!")

    def assign_resource_profile(self, resource_profile_name):
        """
        Assigns Resource Profile
        :param resource_profile_name: resource profile name
        :return:
        """
        if self.get_app_instance_detail():
            scope(self.handle, 'app-instance {} {}'.format(self.app_name, self.identifier))
            self.handle.execute("set resource-profile-name {}".format(resource_profile_name))
            output = self.handle.execute("commit")
            if 'Error' in output:
                logger.info("Failed while trying to assign resource profile with error: {}".format(output))
                raise Exception("Failed while trying to assign resource profile")
            self.handle.execute("discard-buffer")
            output = create_dict(self.handle.execute("show detail"))
            assert output[
                       'Profile Name'] == resource_profile_name, "Mismatch in Profile Name! Expected: {}, Saw: {}".format(
                resource_profile_name, output['Profile Name'])
        else:
            raise Exception("The app is not present. Cannot add resource profile!")

    def is_mi_supported(self):
        """
        receives the self and tests if Multi instance is supported
        ways to check
        "create app-instance" and see if it takes an additional parameter

        scope app asa x.x.x.x and do "show app-info", that command is not supported in the past

        see if scope ssa, "show app" has a column with "Supported Deploy Types"
        """
        self.handle.execute("scope ssa")
        self.handle.execute("scope slot {}".format(self.slot))
        try:
            self.handle.execute("create app-instance ftd testmi")
            self.handle.execute("end")
            self.handle.execute("discard-buffer")
            logger.info("We were able to create app-instance in new format, multi instance is supported")
            self.mi_supported = True
        except Exception:
            raise Exception("MI is not supported")



from ...elektra.actions import Elektra
from ...fmc.actions import Fmc
from ...m3.actions import M3
from ...m4.actions import M4
from ...m5.actions import M5
from ...ftd5500x.actions import Ftd5500x
from ...ssp.actions import Ssp
from ...kp.actions import Kp
from ...series3.actions import Series3
from ...ep.actions import Ep
from ...chassis.actions import Chassis
from ...wm.actions import Wm
from ...asa.actions import Asa

MODEL_TO_HW_MAP = {
    '63': Series3,
    '66': Fmc,
    '69': Elektra,
    '72': Elektra,
    '75': Ftd5500x,
    '76': Ssp,
    '77': Kp,
    'ep': Ep,
    'chassis': Chassis,
    '78': Wm,
    'asa': Asa
}


class Factory():
    @staticmethod
    def factory_by_model(model_number, args=(), kwargs={}):

        """from model number (e.g., '75') and model id (e.g., 'K', return the
        proper device class.

        When invoking function, user must specify args and kwargs to avoid
        confusion.

        :param model_number: string
        :return: the class

        """

        return MODEL_TO_HW_MAP[model_number](*args, **kwargs)

    @staticmethod
    def factory_by_name(class_name, args=(), kwargs={}):
        """from user specified device class name, return the proper device
        class.

        :param class_name: string (has to match the name of device class)
        :return: the class

        """

        return globals()[class_name](*args, **kwargs)

    @staticmethod
    def factory_by_version(device_family, version, args=(), kwargs={}):
        """users tells what family it belongs to (e.g., 'Elektra'), and version
        number (e.g., '96.1(1)47', return the proper device class.

        :param device_family: string
        :param version: string
        :return: the class

        """

        if device_family == 'Elektra':
            if version == '96.1(1)47':
                return Elektra(*args, **kwargs)
            else:
                return Elektra(*args, **kwargs)
        elif device_family == 'Fmc':
            if version.lower() == 'm3':
                return M3(*args, **kwargs)
            elif version.lower() == 'm4':
                return M4(*args, **kwargs)
            elif version.lower() == 'm5':
                return M5(*args, **kwargs)
            else:
                return Fmc(*args, **kwargs)
        else:
            raise RuntimeError('unknown device: {}'.format(device_family))



"""Library to access the webserver."""
import re
import time
import logging
logger = logging.getLogger(__name__)

from unicon.eal.expect import Spawn, TimeoutError
from unicon.utils import AttributeDict
from unicon.eal.dialogs import Dialog

MAX_RETRY_COUNT = 3


class Webserver:
    def __init__(self, hostname='AST-INSTALL', login_username='', login_password='',
                 sudo_password=''):
        """Constructor for Webserver.

        :param hostname: hostname for the device
        :param login_password: device login password for a user
        :param sudo_password: device sudo password for 'root'
        :return: None

        """

        # set hostname and enable_password
        from .constants import WebserverConstants
        WebserverConstants.hostname = hostname
        WebserverConstants.username = login_username
        WebserverConstants.login_password = login_password
        WebserverConstants.sudo_password = sudo_password

        # now that FtdConstants is initialized, create the state machine
        # that contains the proper variables

        from .statemachine import WebserverStatemachine
        self.sm = WebserverStatemachine()

    def ssh_vty(self, ip, port='22'):
        """Setup ssh connection to the server.

        :param ip: ssh host/ip
        :param port: ssh port
        :return: ssh_line

        """
        
        from .constants import WebserverConstants
        spawn = Spawn('ssh -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no '
                      '-l {} -p {} {} \n'.format(WebserverConstants.username, port, ip))
        d1 = Dialog([
            ['continue connecting (yes/no)?', 'sendline(yes)', None, True, False],
            ['(p|P)assword:', 'sendline({})'.format(WebserverConstants.login_password),
             None, False, False],
        ])
        try:
            d1.process(spawn, timeout=60)
            spawn.sendline()
        except TimeoutError:
            pass

        ssh_line = WebserverLine(spawn, self.sm, 'ssh_ftd')
        return ssh_line


class WebserverLine:

    def __init__(self, spawn, sm, type):
        """Constructor for ssh line. Attributes that are None will be set in
        rommon_to_new_image()

        :param spawn: spawn of the line
        :param sm: object of statemachine
        :param type: line type, e.g. 'ssh_vty'
        :return: object of the line

        """

        self.spawn = spawn
        self.sm = sm
        self.type = type
        sm.go_to('any', spawn)

    def go_to(self, state):
        """Go to specified state.

        :param state: string of the state to go to
        :return: None

        """

        self.sm.go_to(state, self.spawn)

    def execute(self, cmd, timeout=10):
        """Run a command, get the console output. Then exclude the command
        echo, and the new prompt to return the content in between.

        :param cmd: command to be executed
        :param timeout: timeout for the command to be executed, defaulted to 10 seconds
        :return: output of the command executed

        """

        current_state = self.sm.current_state
        current_prompt = self.sm.get_state(current_state).pattern

        self.spawn.sendline(cmd)
        # expect and trim cmd
        # commented for long command
        # output_cmd = self.spawn.expect(cmd, timeout=timeout).last_match.string
        # expect and trim current_prompt
        output = self.spawn.expect(current_prompt,
                                   timeout=timeout).last_match.string

        logger.debug("before trimming in execute(): {}".format(output))

        r = re.search(current_prompt, output)
        end = r.span()[0]

        return_data = output[:end].strip()

        return return_data

    def execute_lines(self, cmd_lines, timeout=10):
        """When given a string of multiple lines, run them one by one and
        return the result for the last line.

        :param cmd_lines: commands to be executed line by line
        :param timeout: timeout for individual command to be executed,
                        defaulted to 10 seconds
        :return: output of the last command executed

        """

        for cmd in cmd_lines.split('\n'):
            cmd = cmd.strip()
            if cmd == "":
                continue # empty line
            return_data = self.execute(cmd, timeout=timeout)

        return return_data

    def disconnect(self):
        """Disconnect the session.

        :return: None

        """

        if self.type == 'ssh_vty':
            self.sm.go_to('user_state', self.spawn)
            self.spawn.sendline('exit')
            self.spawn.expect('closed.')
            logger.info('ssh to ftd connection disconnected')




import logging
import subprocess

from unicon.eal.dialogs import Dialog
from unicon.utils import AttributeDict

try:
    import kick.graphite.graphite as graphite
except ImportError:
    import kick.metrics.metrics as graphite

from .patterns import EpPatterns
from .statemachine import EpStatemachine
from ...general.actions.basic import BasicDevice, BasicLine, NewSpawn

logger = logging.getLogger(__name__)

DEFAULT_TIMEOUT = 10


class Ep(BasicDevice):
    def __init__(self, hostname='firepower',
                 login_username='admin',
                 login_password='Admin123',
                 sudo_password='Admin123',
                 *args,
                 **kwargs):
        """Constructor for Ep.

        :param hostname: host name in prompt e.g. FS2000-01
        :param login_username: user name for login
        :param login_password: password for login
        :param ep_root_password: root password for EP
        :return: None

        Add *args and **kwargs so that the initiator can be invoked with
        additional (unused) arguments.

        """
        graphite.publish_kick_metric('device.ep.init', 1)
        self.patterns = EpPatterns(
            hostname=hostname,
            login_username=login_username,
            login_password=login_password,
            sudo_password=sudo_password)
        self.sm = EpStatemachine(self.patterns)
        self.line_class = EpLine

        self.set_default_timeout(DEFAULT_TIMEOUT)

    def ssh_vty(self, ip, port, username='admin', password='Admin123',
                timeout=None, line_type='ssh', rsa_key=None):
        """Set up an ssh connection to device's interface.

        This goes into device's ip address, not console.

        :param ip: ip address of terminal server
        :param port: port of device on terminal server
        :param username: usually "admin"
        :param password: usually "Admin123"
        :param line_type: ssh line type
        :param timeout: in seconds
        :param rsa_key: identity file (full path)
        :return: a line object (where users can call execute(), for example)

        """

        graphite.publish_kick_metric('device.basic.ssh_vty', 1)
        if not timeout:
            timeout = self.default_timeout

        if rsa_key:
            resp = subprocess.getoutput('chmod 400 {}'.format(rsa_key))
            if 'No such file or directory' in resp:
                raise RuntimeError('The identity file {} you provided does not exist'.format(rsa_key))
            spawn_id = NewSpawn('ssh -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no '
                                '-i {} -l {} -p {} {} \n'.format(rsa_key, username, port, ip))
        else:
            spawn_id = NewSpawn('ssh -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no '
                                '-l {} -p {} {} \n'.format(username, port, ip))

        ctx = AttributeDict({'password': password})
        d = Dialog([
            ['continue connecting (yes/no)?', 'sendline(yes)', None, True, False],
            ['(p|P)assword:', 'sendline_ctx(password)', None, True, False],
            ['Password OK', 'sendline()', None, False, False],
            ['[.*>#$]+.*\s', 'sendline()', None, False, False],
        ])
        d.process(spawn_id, context=ctx, timeout=timeout)

        ssh_line = self.line_class(spawn_id, self.sm, line_type, timeout)
        logger.debug('ssh_vty() finished successfully')

        return ssh_line


class EpLine(BasicLine):
    def __init__(self, spawn_id, sm, type, timeout):
        """ Constructor of EpLine

                :param spawn_id: spawn_id of the line
                :param sm: object of statemachine
                :param type: line type, e.g. 'ssh', 'telnet'
                :param timeout: in seconds

                :return: None

                """
        super().__init__(spawn_id=spawn_id, sm=sm, type=type, timeout=timeout)
        self.default_timeout = timeout

    def go_to(self, state, timeout=None):
        """Override parent go_to function to enable hop_wise flag."""

        if not timeout:
            timeout = self.default_timeout

        super().go_to(state, hop_wise=True, timeout=timeout)

    def sudo_execute(self, cmd, timeout=None, exception_on_bad_command=False):
        self.go_to('sudo_state')
        return self.execute(cmd, timeout, exception_on_bad_command)

    def expert_execute(self, cmd, timeout=None, exception_on_bad_command=False):
        self.go_to('admin_state')
        return self.execute(cmd, timeout, exception_on_bad_command)

    def replace_asa_image(self, source_location, pwd, timeout=300):
        """ Not implemented for EpLine class"""

        raise NotImplementedError



from unicon.eal.dialogs import Statement


def password_handler(spawn, password):
    spawn.sendline(password)
    spawn.sendline()


def enable_handler(spawn, password):
    spawn.sendline('enable')
    spawn.expect('Password: ')
    spawn.sendline(password)


class ElektraStatements:
    def __init__(self, patterns):
        self.login_password_statement = Statement(pattern=patterns.prompt.password_prompt,
                                                  action=password_handler,
                                                  args={'password': patterns.login_password},
                                                  loop_continue=True,
                                                  continue_timer=True)

        self.sudo_password_statement = Statement(pattern=patterns.prompt.password_prompt,
                                                 action=password_handler,
                                                 args={'password': patterns.sudo_password},
                                                 loop_continue=True,
                                                 continue_timer=True)

        self.enable_password_statement = Statement(pattern=patterns.prompt.disable_prompt,
                                                   action=enable_handler,
                                                   args={'password': patterns.enable_password},
                                                   loop_continue=True,
                                                   continue_timer=True)

        self.expert_password_statement = Statement(pattern=patterns.prompt.password_prompt,
                                                  action=password_handler,
                                                  args={'password': patterns.default_password},
                                                  loop_continue=True,
                                                  continue_timer=True)




from unicon.statemachine import State, Path, StateMachine
from unicon.eal.dialogs import Dialog
from .patterns import Elektra5585Patterns
from .statements import Elektra5585Statements
from .dialogs import Elektra5585Dialog


class Elektra5585Statemachine(StateMachine):
    def __init__(self, patterns):
        """Initializer of Elektra5585StateMachine."""
        self.patterns = patterns
        self.dialogs = Elektra5585Dialog(patterns)
        self.statements = Elektra5585Statements(patterns)

        super().__init__()

    def create(self):
        # Create States
        prelogin_state = State('prelogin_state', self.patterns.prompt.prelogin_prompt)
        pkglogin_state = State('pkglogin_state', self.patterns.prompt.pkglogin_prompt)
        fireos_state = State('fireos_state', self.patterns.prompt.fireos_prompt)
        expert_state = State('expert_state', self.patterns.prompt.expert_prompt)
        sudo_state = State('sudo_state', self.patterns.prompt.sudo_prompt)
        rommon_state = State('rommon_state', self.patterns.prompt.rommon_prompt)
        boot_state = State('boot_state', self.patterns.prompt.firepower_boot)
        config_state = State('config_state', self.patterns.prompt.config_prompt)
        system_state = State('system_state', self.patterns.prompt.system_prompt)

        # Add states
        self.add_state(prelogin_state)
        self.add_state(pkglogin_state)
        self.add_state(fireos_state)
        self.add_state(expert_state)
        self.add_state(sudo_state)
        self.add_state(rommon_state)
        self.add_state(boot_state)
        self.add_state(config_state)
        self.add_state(system_state)

        # Create paths
        prelogin_to_fireos_path = Path(prelogin_state, fireos_state, 'admin',
                                       Dialog([self.statements.login_password_statement]))
        pkglogin_to_boot_path = Path(pkglogin_state, boot_state, 'admin',
                                       Dialog([self.statements.pkg_password_statement]))
        fireos_to_prelogin_path = Path(fireos_state, prelogin_state, 'exit', None)
        fireos_to_expert_path = Path(fireos_state, expert_state, 'expert', None)
        expert_to_fireos_path = Path(expert_state, fireos_state, 'exit', None)
        expert_to_sudo_path = Path(expert_state, sudo_state, 'sudo su -',
                                   Dialog([self.statements.sudo_password_statement]))
        sudo_to_expert_path = Path(sudo_state, expert_state, 'exit', None)
        system_to_expert_path = Path(system_state, expert_state, 'expert', None)
        expert_to_system_path = Path(expert_state, system_state, 'exit', None)

        # Add paths
        self.add_path(prelogin_to_fireos_path)
        self.add_path(pkglogin_to_boot_path)
        self.add_path(fireos_to_prelogin_path)
        self.add_path(fireos_to_expert_path)
        self.add_path(expert_to_fireos_path)
        self.add_path(expert_to_sudo_path)
        self.add_path(sudo_to_expert_path)
        self.add_path(system_to_expert_path)
        self.add_path(expert_to_system_path)
